//
//  aria - yet another download tool
//  Copyright (C) 2000, 2001 Tatsuhiro Tsujikawa
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

// $Id: ItemCell.cc,v 1.36 2001/03/22 08:42:17 tujikawa Exp $

// class ItemCell implementation
#include "ItemCell.h"
#include "ItemStatusDynamic.h"
#include "ItemStatusLog.h"
#include "ItemStatusPartial.h"
#include "ItemStatusRec.h"

extern ItemOption *itemoption;
extern void Send_report(MessageType reporttype, void *data);
extern bool CRC_check_main(ItemCell *itemcell);
extern pthread_mutex_t itemlistlock;

ItemCell::ItemCell(const string& url_in,
		   const URLcontainer& urlcon_in,
		   const Options& options_in,
		   const string& initial_log)
{
  id = -1;
  urlcon = urlcon_in;
  options = options_in;
  // set username and password
  if(urlcon.ret_Username().size()) {
    options.set_use_authentication(true);
    Userdata userdata(urlcon.ret_Username(), urlcon.ret_Password());
    options.set_userpasswd(userdata);
  }
  url = urlcon.ret_URL();

  // initialize log message list

  loglist.push_back(ItemLogCell(initial_log, MSG_INIT));

  msgpipe[0] = -1;
  msgpipe[1] = -1;

  crc = 0;
  crc_type = CRC_NONE;
  status = ITEM_READY;
  currentsize = 0;
  totalsize = 0;
  count = 0;

  url_opt = url;
  urlcon_opt = urlcon;
  options_opt = options;

  flag_opt_updated = false;
  split_complete = true;
  // initialize mutex locks
  pthread_mutex_init(&itemloglock, NULL);
  pthread_mutex_init(&optionslock, NULL);
  pthread_mutex_init(&crclock, NULL);
  // init logging option
  log_flag = true;
  // init error code
  item_errno = ITEM_ENONE;
  // init event type
  dl_event = ItemCommand::EV_NOEVENT;
}

ItemCell::ItemCell()
{
  split_complete = true;
  count = 0;

  msgpipe[0] = -1;
  msgpipe[1] = -1;

  // initialize mutex locks
  pthread_mutex_init(&itemloglock, NULL);
  pthread_mutex_init(&optionslock, NULL);
  pthread_mutex_init(&crclock, NULL);
  // init event type
  dl_event = ItemCommand::EV_NOEVENT;
}

ItemCell::~ItemCell()
{
  if(msgpipe[0] > 0) close(msgpipe[0]);
  if(msgpipe[1] > 0) close(msgpipe[1]);
  pthread_mutex_destroy(&itemloglock);
  pthread_mutex_destroy(&optionslock);
}

const ItemLogList& ItemCell::ret_Log_list() const
{
  return loglist;
}

void ItemCell::Clear_log()
{
  pthread_mutex_lock(&itemloglock);
  loglist.clear();
  pthread_mutex_unlock(&itemloglock);
}

void ItemCell::reset_Count()
{
  count = 0;
}

void ItemCell::inc_Count()
{
  ++count;
}

unsigned int ItemCell::ret_Count() const
{
  return count;
}

void ItemCell::begin_split()
{
  split_complete = false;
}

void ItemCell::end_split()
{
  split_complete = true;
}

void ItemCell::set_logging(bool flag)
{
  log_flag = flag;
}

void ItemCell::Send_message_to_gui(const string& message,
				   MessageType reporttype)
{
  pthread_mutex_lock(&itemloglock);
  ItemLogCell itemlogcell(message, reporttype);
  if(log_flag) Append_itemlog(itemlogcell);
  pthread_mutex_unlock(&itemloglock);
  ItemStatusLog *itemstatus = new ItemStatusLog(this, itemlogcell);

  Send_report(reporttype, itemstatus);
}

void ItemCell::Send_partial(ItemCell *itemcell_partial)
{
  ItemStatusPartial *itemstatus = new ItemStatusPartial(this, itemcell_partial);
  Send_report(MSG_DOWNLOAD_ADD_PARTIAL, itemstatus);
}

void ItemCell::Send_status()
{
  ItemStatusDynamic *itemstatus = new ItemStatusDynamic(this, status, count, currentsize, totalsize);
  Send_report(MSG_DOWNLOAD_STATUS, itemstatus);
}

void ItemCell::Send_status(float speed)
{
  ItemStatusDynamic *itemstatus = new ItemStatusDynamic(this, status, count, currentsize, totalsize, speed);
  Send_report(MSG_DOWNLOAD_STATUS, itemstatus);
}

void ItemCell::Send_status_complete()
{
  ItemStatusDynamic *itemstatus = new ItemStatusDynamic(this, status, count, currentsize, totalsize);
  itemstatus->set_DeleteFlag(true);
  Send_report(MSG_DOWNLOAD_STATUS, itemstatus);
} 

void ItemCell::SEND(const string& command, const Socket& socket)
{
  list<int> fd_list;

  int retval = socket.is_writeready(0, fd_list);
  if(retval && socket.is_set(fd_list)) {
    // send GET request to WWW server
    if(socket.Send(command) < 0) {
      // failed to send request
      throw ITEM_ESEND;
    } else {
      // send successful;
    }
  } else {
    // cannot send data because of error
    throw ITEM_ESEND;
  }
}

unsigned int ItemCell::Get_starting_byte()
{
  struct stat filestat;

  string filename;
  //if(documentroot_dir == options.ret_Store_Dir() && options.ret_recurse_count() > 1 && options.ret_with_hostname_dir()) {
  //filename = options.ret_Store_Dir()+urlcon.ret_Hostname()+urlcon.ret_Dir()+'/'+ret_Filename();
  // } else {
  filename = options.ret_Store_Dir()+ret_Filename();
  //}
  // rollback featrue added 2001/3/17
  if(stat(filename.c_str(), &filestat) < 0 || !S_ISREG(filestat.st_mode) ||
     filestat.st_size <= options.ret_Rollback_bytes()) {
    return 0;
  } else {
    if(truncate(filename.c_str(), filestat.st_size-options.ret_Rollback_bytes()) < 0) {
      Send_message_to_gui(_("unable to truncate the file '")+filename+_("'; rollback disabled"), MSG_DOWNLOAD_ERROR);

      return filestat.st_size;
    } else {
      return filestat.st_size-options.ret_Rollback_bytes();
    }
  }
}

ItemCell::DownloadStatusType ItemCell::Process_command(ItemCommand itemcommand)
{
  switch(itemcommand.commandtype) {
  case ItemCommand::COMMAND_DELETE:
    return DLDELETE;
  case ItemCommand::COMMAND_HALT:
    return DLHALT;
  case ItemCommand::COMMAND_DOWNLOAD_AGAIN:
    {
      // delete file
      string filename = options.ret_Store_Dir()+ret_Filename();
      unlink(filename.c_str());
      
      Send_message_to_gui(_("download again"), MSG_DOWNLOAD_INFO);
      set_Status(ITEM_DOWNLOAD_AGAIN);
      return DLAGAIN;
    }
  case ItemCommand::COMMAND_CHANGE_SPEED:
    get_Options_Lock();
    options_opt.set_speed_limit(itemcommand.value);
    options.set_speed_limit(itemcommand.value);
    release_Options_Lock();
    return DLCHANGE;
  case ItemCommand::COMMAND_STOP:
  default:
    return DLSTOP;
  }
}

ItemCell::DownloadStatusType ItemCell::ret_Dl_status() const
{
  return dl_status;
}

ItemCommand::EventCause ItemCell::ret_Dl_event() const
{
  return dl_event;
}

void ItemCell::Show_option_window()
{
  itemoption->set_Option_Values(this, options_opt);
  itemoption->Show_Option_Window();
}

// inherited class must override this method
ItemCell::DownloadStatusType ItemCell::Download_Main()
{
  cerr << "in Download_Main()" << endl;
  cerr << "inherited class must override this method" << endl;
  return DLSUCCESS;
}


ItemCell::CRC_Type ItemCell::ret_CRC_Type() const
{
  return crc_type;
}

const string& ItemCell::ret_URL() const
{
  return url;
}

unsigned int ItemCell::ret_CRC() const
{
  return crc;
}

ItemCell::ItemStatusType ItemCell::ret_Status() const
{
  return status;
}

unsigned int ItemCell::ret_Size_Current() const
{
  return currentsize;
}

unsigned int ItemCell::ret_Size_Total() const
{
  return totalsize;
}

int ItemCell::ret_Desc_w() const
{
  return msgpipe[1];
}

int ItemCell::ret_Desc_r() const
{
  return msgpipe[0];
}

string ItemCell::ret_Filename() const
{
  // because ulcon.file returns something like "/filename",
  // so we need to ignore first character '/'
  //string filename = urlcon.ret_File();

  //int slash_index = filename.rfind('/');
  //if(slash_index >= 0) {
  //  filename.erase(0, slash_index+1);
  //}

  return urlcon.ret_Filename();
}

string ItemCell::ret_Filename_opt() const
{
  // because ulcon.file returns something like "/filename",
  // so we need to ignore first character '/'
  //string filename = urlcon_opt.ret_File();

  //int slash_index = filename.rfind('/');
  //if(slash_index >= 0) {
  //  filename.erase(0, slash_index+1);
  //}

  return urlcon_opt.ret_Filename();
}

URLcontainer& ItemCell::ret_URL_Container()
{
  return urlcon;
}

void ItemCell::set_CRC_Type(CRC_Type crc_type_new)
{
  crc_type = crc_type_new;
}

void ItemCell::set_CRC(unsigned int crc_new)
{
  crc = crc_new;
}

void ItemCell::set_Status(ItemStatusType status_new)
{
  status = status_new;
}

void ItemCell::set_Size_Current(unsigned int currentsize_new)
{
  currentsize = currentsize_new;
}

void ItemCell::set_Size_Total(unsigned int totalsize_new)
{
  totalsize = totalsize_new;
}

void ItemCell::set_URL_Container(const URLcontainer& urlcon_new)
{
  urlcon = urlcon_new;
}

void ItemCell::set_URL(const string& url_in)
{
  url = url_in;
}

void ItemCell::Append_itemlog(const ItemLogCell& itemlogcell)
{
  loglist.push_back(itemlogcell);
}

void ItemCell::Open_Desc()
{
  pipe(msgpipe);
}

void ItemCell::Close_Desc()
{
  close(msgpipe[0]);
  close(msgpipe[1]);
  msgpipe[0] = -1;
  msgpipe[1] = -1;
}

// some mutex locks
void ItemCell::get_Log_Lock()
{
  pthread_mutex_lock(&itemloglock);
}

void ItemCell::release_Log_Lock()
{
  pthread_mutex_unlock(&itemloglock);
}

void ItemCell::get_Options_Lock()
{
  pthread_mutex_lock(&optionslock);
}

void ItemCell::release_Options_Lock()
{
  pthread_mutex_unlock(&optionslock);
}

void ItemCell::get_CRC_Lock()
{
  pthread_mutex_lock(&crclock);
}

void ItemCell::release_CRC_Lock()
{
  pthread_mutex_unlock(&crclock);
}

void ItemCell::set_Options(const Options& options_in)
{
  options = options_in;
}

Options& ItemCell::ret_Options()
{
  return options;
}

void ItemCell::set_Errno(ItemErrorType item_errno_in)
{
  item_errno = item_errno_in;
}

ItemCell::ItemErrorType ItemCell::ret_Errno() const
{
  return item_errno;
}

void ItemCell::set_Command(DownloadStatusType dl_status_in)
{
  dl_status = dl_status_in;
  dl_event = ItemCommand::EV_INTERNAL; //internal event
}

void ItemCell::set_Command(const ItemCommand& itemcommand_in)
{
  dl_status = Process_command(itemcommand_in); // convert
  dl_event = itemcommand_in.eventtype;
}

ItemCell::DownloadStatusType ItemCell::ret_Command() const
{
  return dl_status;
}

bool ItemCell::Is_Partial() const
{
  return false;
}

void ItemCell::PERROR(ItemErrorType item_errno)
{
  string report;

  switch(item_errno) {
  case ITEM_ECANTRESOLVE:
    Send_message_to_gui(_("can't resolve host"), MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ESOCKET:
    Send_message_to_gui(_("socket error"), MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ECONNREFUSED:
    report = _("connection refused - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ETIMEDOUT:
    Send_message_to_gui(_("time out"), MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EIO:
    report = _("I/O error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EIOFILE:
    report = _("FILE I/O error - maybe wrong save location or disk full?");
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ELOCATION:
    Send_message_to_gui(_("Location retrieval failed"), MSG_DOWNLOAD_ERROR);
    //Send_message_to_gui(_("Aria cannot handle redirection to FTP directory. Sorry"), MSG_DOWNLOAD_ERROR); 
    break;
  case ITEM_ECANTFINDURL:
    Send_message_to_gui(_("cannot find embedded URL"), MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EFTPCOM:
    Send_message_to_gui(_("FTP server returned negative response"), MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EBIND:
    report = _("bind error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ELISTEN:
    report = _("listen error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EACCEPT:
    report = _("accept error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ESEND:
    report = _("send error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_ERECV:
    report = _("recv error - ")+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  case ITEM_EPROT:
    report = _("general protocol error");//+itos(errno)+" : "+strerror(errno);
    Send_message_to_gui(report, MSG_DOWNLOAD_ERROR);
    break;
  default:
    Send_message_to_gui(_("some error occured"), MSG_DOWNLOAD_ERROR);
    break;
  }
}

// this function is originary appeared in SOCKET PROMGAMMING FAQ
int ItemCell::inet_hostaddr(const string& hostname, struct in_addr* addr)
{
    struct hostent *host;

    //#ifdef HAVE_INET_ATON
    if(inet_aton(hostname.c_str(),addr)) {
      //#else
      //if(inet_pton(AF_INET,hostname.c_str(),addr) > 0) {
      //#endif
      return 1;
    }

    host = gethostbyname(hostname.c_str());
    if(!host) {
      return 0;
    }
    *addr = *(struct in_addr*)(host->h_addr_list[0]);
    return 1;
}

void ItemCell::Make_TCP_connection(const Socket& socket, unsigned int port_in)
{
  Make_TCP_connection(socket, urlcon.ret_Hostname(), port_in);
}

void ItemCell::Make_TCP_connection(const Socket& socket, const string& server, unsigned int port_in)
{
  sockaddr_in serv_addr;

  memset((char *)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;

  string hostname;
  unsigned port;
  if(ret_URL_Container().ret_Protocol() == "http:" && options.ret_use_http_proxy() && !options.ret_http_proxy().ret_Server().empty()) {

    hostname = options.ret_http_proxy().ret_Server();
    port = options.ret_http_proxy().ret_Port();
  } else {
    hostname = server;//ret_URL_Container().ret_Hostname();
    port = port_in;
  }
  string report = _("connecting to ")+hostname+":"+itos(port);
  Send_message_to_gui(report, MSG_DOWNLOAD_INFO);

  serv_addr.sin_port = htons(port);
  serv_addr.sin_family = AF_INET;
  if(!inet_hostaddr(hostname, &serv_addr.sin_addr) ||
     !serv_addr.sin_port) {
    throw ITEM_ECANTRESOLVE;
  }
  // set the socket nonblocking mode
  int flags = socket.Getflags();
  socket.Setflags(flags|O_NONBLOCK);

  // connect to server
  int retval = socket.Connect(serv_addr);
  if(retval < 0 && errno != EINPROGRESS) {
    throw ITEM_ECONNREFUSED;
  }
  unsigned int timedout = ret_Options().ret_Timed_Out();

  list<int> rfd_list;
  list<int> wfd_list;
  rfd_list.push_back(ret_Desc_r());

  retval = socket.is_readwriteready(timedout, rfd_list, wfd_list);
  if(retval && Socket::is_set(ret_Desc_r(), rfd_list)) {
    ItemCommand itemcommand;
    read(ret_Desc_r(), &itemcommand, sizeof(ItemCommand));
    //set_Command(Process_command(itemcommand));
    set_Command(itemcommand);
    if(ret_Dl_status() == DLCHANGE) {
      throw ITEM_EINTER;
    }
  }
  if(retval > 0) {
  } else if(retval == 0) { // timed out
    throw ITEM_ETIMEDOUT;
  } else {     // Connection Refused
    throw ITEM_ECONNREFUSED;
  }

  if(socket.check_error()) {
    throw ITEM_ECONNREFUSED;
  }
  // set the socket blocking mode
  socket.Setflags(flags&~O_NONBLOCK);

  Send_message_to_gui(_("connection established"), MSG_DOWNLOAD_INFO);
}

ItemCell::DownloadStatusType ItemCell::Post_process()
{
  DownloadStatusType retval = DLSUCCESS;
  try {
    if(!options.ret_no_crc_checking() && ret_CRC_Type() != CRC_NONE) {
      if(!CRC_check_main(this)) {
	if(options.ret_ignore_crc_error()) {
	  Send_message_to_gui(_("ignore CRC error"), MSG_DOWNLOAD_INFO);
	} else {
	  retval = DLERROR;
	}
      }
    }
  }
  catch(ItemErrorType err) {
    PERROR(err);
    retval = DLERROR;
  }

  return retval;
}

void ItemCell::Append_worker(ItemCell *itemcell_partial)
{
  worker_list.push_back(itemcell_partial);
}

void ItemCell::Remove_worker(ItemCell *itemcell_partial)
{
  worker_list.remove(itemcell_partial);
}

void ItemCell::Clear_worker()
{
  worker_list.clear();
}

bool ItemCell::No_more_worker() const
{
  if(worker_list.size() || !split_complete) return false;
  else return true;
}

const list<ItemCell*>& ItemCell::ret_Worker_list() const
{
  return worker_list;
}

Options& ItemCell::ret_Options_opt()
{
  return options_opt;
}

void ItemCell::set_Options_opt(const Options& options_in)
{
  options_opt = options_in;
}

void ItemCell::set_URL_opt(const string& url_in)
{
  url_opt = url_in;
}

const string& ItemCell::ret_URL_opt() const
{
  return url_opt;
}

URLcontainer& ItemCell::ret_URL_Container_opt()
{
  return urlcon_opt;
}

void ItemCell::set_URL_Container_opt(const URLcontainer& urlcon_in)
{
  urlcon_opt = urlcon_in;
}

void ItemCell::Raise_option_update_flag()
{
  flag_opt_updated = true;
}

bool ItemCell::Is_option_updated()
{
  return flag_opt_updated;
}

void ItemCell::Apply_new_options()
{
  if(flag_opt_updated) {
    flag_opt_updated = false;
    //options_opt.set_documentroot_dir(options.ret_documentroot_dir());
    if(options.ret_Store_Dir() != options_opt.ret_Store_Dir()) {
      root_url = "";
      documentroot_dir = "";
    }
    options = options_opt;
    urlcon = urlcon_opt;
    url = url_opt;
    count = 0; // reset retry count; added 2001/3/14
    //set_Size_Total(0);
    //set_Size_Current(0);
    //set_Status(ITEM_READY);
    Send_status();
  }
}

#define PCSPLIT 10
unsigned int ItemCell::Download_data(ofstream& outfile, const Socket& socket)
{
  const int databuf_size_def = 4096;
  int databuf_size;
  get_Options_Lock();
  if(databuf_size_def < options_opt.ret_speed_limit()*1024/PCSPLIT || options_opt.ret_speed_limit() < 0.1) {
    databuf_size = databuf_size_def;
  } else {
    databuf_size = (int)(options_opt.ret_speed_limit()*1024)/PCSPLIT;
  }
  release_Options_Lock();
  char databuf[databuf_size_def];
  const int writebuf_size = 12288;
  char *writebuf = new char[writebuf_size];
  char *writebuf_tail = writebuf;
  int size;
  string line;
  struct timeval start_time_, end_time_;
  struct timezone tz_dummy;
  unsigned int timedout = options.ret_Timed_Out();
  unsigned int elapsed_size = 0;
  float speed = 0.0;
  list<int> fd_list;
  unsigned int startsize = currentsize;

  gettimeofday(&start_time_, &tz_dummy);
  Send_status();

  try {
    while(1) {
      fd_list.clear();
      fd_list.push_back(ret_Desc_r());
      int retval = socket.is_readready(timedout, fd_list);
      if(retval && Socket::is_set(ret_Desc_r(), fd_list)) {
	ItemCommand itemcommand;

	read(ret_Desc_r(), &itemcommand, sizeof(ItemCommand));

	//set_Command(Process_command(itemcommand));
	set_Command(itemcommand);
	if(ret_Dl_status() != DLCHANGE) {
	  outfile.write(writebuf, writebuf_tail-writebuf);
	  throw ITEM_EINTER;
	} else {
	  // change buffer size;
	  if(databuf_size_def < options_opt.ret_speed_limit()*1024/PCSPLIT || options_opt.ret_speed_limit() < 0.1 ) {
	    databuf_size = databuf_size_def;
	  } else {
	    //cerr << "bufsize" << (int)(options_opt.ret_speed_limit()*1024)/PCSPLIT << endl;
	    databuf_size = (int)(options_opt.ret_speed_limit()*1024)/PCSPLIT;
	  }
	}
      } else if(retval && socket.is_set(fd_list)) {
	size = socket.Recv(databuf, databuf_size);//sizeof(databuf));
	if(size < 0) {
	  throw ITEM_EIO;
	}

	currentsize += size;
	gettimeofday(&end_time_, &tz_dummy);
	elapsed_size += size;
	long int elapsed = (end_time_.tv_sec-start_time_.tv_sec)*1000000+end_time_.tv_usec-start_time_.tv_usec;

	if((totalsize != 0 && currentsize >= totalsize) || size == 0) {
	  if(Is_Partial()) {
	    unsigned int diff = currentsize-totalsize;
	    memcpy(writebuf_tail, databuf, size-diff);
	    writebuf_tail += size-diff;
	    currentsize = totalsize;
	  } else {
	    memcpy(writebuf_tail, databuf, size);
	    writebuf_tail += size;
	  }
	  Send_status(speed);
	  outfile.write(writebuf, writebuf_tail-writebuf);
	  writebuf_tail = writebuf;
  
	  if(outfile.bad()) {
	    throw ITEM_EIOFILE;
	  }
	  
	  break;
	} else {
	  if(elapsed > 1000000) {
	    speed = ((elapsed_size/1024.0)/(elapsed/1000000.0)+speed)/2;
	    Send_status(speed);
	    gettimeofday(&start_time_, &tz_dummy);
	    elapsed_size = 0;
	  } else {
	    //cerr << elapsed_size << endl;
	    if(options_opt.ret_speed_limit() > 0.0 &&
	       elapsed_size >= (unsigned int)(options_opt.ret_speed_limit()*1024*0.8)) {  	       
	      struct timeval tv;
	      tv.tv_sec = 0;
	      tv.tv_usec = 1000000-elapsed;  
	      select(ret_Desc_r()+1, NULL, NULL, NULL, &tv);
	    }
	  }
	  memcpy(writebuf_tail, databuf, size);
	  writebuf_tail += size;
	  if(writebuf_size-(writebuf_tail-writebuf) < databuf_size+databuf_size_def) {
	    outfile.write(writebuf, writebuf_tail-writebuf);
	    writebuf_tail = writebuf;
	  }
	  if(outfile.bad()) {
	    throw ITEM_EIOFILE;
	  }
	}
	
      } else {
	// timed out
	outfile.write(writebuf, writebuf_tail-writebuf);// buffer flush
	throw ITEM_ETIMEDOUT;
      }
    }
  } catch (ItemErrorType err) {
    delete [] writebuf;
    throw err;
  }
  delete [] writebuf;
  return currentsize-startsize;
}


bool ItemCell::Concatenate_partial_files()
{
  try {
    string filename = options.ret_Store_Dir()+ret_Filename();
    //string partial_filename = filename+".-1";
    //struct stat file_stat;
    unsigned int index_start;
    /*
    if(stat(partial_filename.c_str(), &file_stat) == -1 || !S_ISREG(file_stat.st_mode)) {
      partial_filename = filename+".0";
      if(rename(partial_filename.c_str(), filename.c_str())) {
	throw ITEM_EIO;
      }
      index_start = 1;
    } else {
      if(rename(partial_filename.c_str(), filename.c_str())) {
	throw ITEM_EIO;
      }
      index_start = 0;
    }
    */
    string partial_filename = filename+".0";
    if(rename(partial_filename.c_str(), filename.c_str())) {
      throw ITEM_EIO;
    }
    index_start = 1;
    string indexfilename = partial_filename+".index";
    unlink(indexfilename.c_str());

    ofstream outfile(filename.c_str(), ios::binary|ios::out|ios::app);
    if(outfile.bad()) {
      throw ITEM_EIO;
    }
    for(unsigned int i = index_start; i < options.ret_Divide(); ++i) {
      ifstream infile;
      partial_filename = filename+"."+itos(i);
      infile.open(partial_filename.c_str(), ios::binary|ios::in);
      if(infile.bad()) {
	throw ITEM_EIO;
      }
      while(!infile.eof()) {
	char buf[8192];
	infile.read(buf, sizeof(buf));
	outfile.write(buf, infile.gcount());

	if(infile.bad() || outfile.bad()) {
	  throw ITEM_EIO;
	}
      }
      infile.close();
      unlink(partial_filename.c_str());
      string indexfilename = partial_filename+".index";
      unlink(indexfilename.c_str());
    }
    outfile.close();

    /*
    FILE *outfile = fopen(filename.c_str(), "a");
    if(outfile == NULL) throw ITEM_EIO;
    for(unsigned int i = index_start; i < options.ret_Divide(); ++i) {
      partial_filename = filename+"."+itos(i);
      FILE *infile = fopen(partial_filename.c_str(), "r");
      if(infile == NULL) {
	throw ITEM_EIO;
      }
      while(1) {
	char buf[8192];
	size_t size = fread(buf, sizeof(buf), 1, infile);
	if(size == 0) {
	  if(ferror(infile)) throw ITEM_EIO;
	  break;
	}
	size_t size_w = fwrite(buf, sizeof(buf), 1, outfile);
	if(size_w < size) {
	  throw ITEM_EIO;
	}
      }
      fclose(infile);
      unlink(partial_filename.c_str());
    }
    fclose(outfile);
    */
    return true;
  }
  catch(ItemErrorType err) {
    PERROR(err);
    return false;
  }
}

void ItemCell::set_id(int id_in)
{
  //if(id != -1 && id_in != -1) cerr << "warning: new id provided, while old id is still valid" << endl;
  id = id_in;
}

int ItemCell::ret_id() const
{
  return id;
}

const string& ItemCell::ret_root_url() const
{
  return root_url;
}

void ItemCell::set_root_url(const string& root_url_in)
{
  root_url = root_url_in;
}

const string& ItemCell::ret_documentroot_dir() const
{
  return documentroot_dir;
}

void ItemCell::set_documentroot_dir(const string& documentroot_dir_in)
{
  documentroot_dir = documentroot_dir_in;
}

void ItemCell::Send_status_recursive(const list<ItemCell *>& item_list)
{
  ItemStatusRec *itemstatus = new ItemStatusRec(this, item_list);
  Send_report(MSG_DOWNLOAD_STATUS, itemstatus);
}

void ItemCell::Make_directory_if_needed()
{
  pthread_mutex_lock(&itemlistlock);
  struct stat dir_stat;
  if(stat(options.ret_Store_Dir().c_str(), &dir_stat) == -1 ||
     !S_ISDIR(dir_stat.st_mode)) {
    string temp_storedir = options.ret_Store_Dir();

    if(temp_storedir.at(temp_storedir.size()-1) == '/') {
      temp_storedir.erase(temp_storedir.size()-1);
    }
    string dir_to_make = Token_splitter(temp_storedir, "/");
    
    while(temp_storedir.size()) {
      dir_to_make += '/'+Token_splitter(temp_storedir, "/");
      struct stat dir_stat;
      if(stat(dir_to_make.c_str(), &dir_stat) == -1 || !S_ISDIR(dir_stat.st_mode)) {
	if(mkdir(dir_to_make.c_str(), 0755) < 0) {
	  Send_message_to_gui(_("faild to make directory '")+dir_to_make+"'\nreason:"+strerror(errno), MSG_DOWNLOAD_ERROR);
	  set_Errno(ITEM_EIOFILE);
	  pthread_mutex_unlock(&itemlistlock);
	  throw ITEM_EIOFILE;
	}
      }
    }
  }
  pthread_mutex_unlock(&itemlistlock);
}

bool ItemCell::Is_older_than_remote(time_t modtime)
{
  string filename = options.ret_Store_Dir()+ret_Filename();
  struct stat filestat;
  //cerr << filename << endl;
  if(stat(filename.c_str(), &filestat) < 0 || !S_ISREG(filestat.st_mode)) {
    return false;
  } else if(filestat.st_mtime > modtime) return false;
  else return true;
}
